
#define GLM_ENABLE_EXPERIMENTAL
#include<glm/gtx/hash.hpp>
#include<unordered_map>
#include<algorithm>
#include<set>
#include<functional>

struct Vertex {
	vec3 pos;
	vec2 texCoord;

	bool operator==(const Vertex& other) const {
		return pos == other.pos && texCoord == other.texCoord;
	}
};// Can't use  __attribute__((packed)) due to non-POD vec2 and vec3

namespace std {
	template<> struct hash<Vertex> {
		size_t operator()(Vertex const& vertex) const {
			return ((hash<glm::vec3>()(vertex.pos) ^ 
						(hash<glm::vec2>()(vertex.texCoord) << 1)));
		}
	};
}

unsigned int generic_program = -1;

struct GameObject {
	unsigned int vbuf, ebuf, tbuf, lbuf, vtbuf;
	unsigned int vertex_shader;
	unsigned int fragment_shader;
	unsigned int program;
	unsigned int vertex_attrib; 
	unsigned int tex_attrib;
	unsigned int mvp_uniform;
	unsigned int tex_uniform;
	unsigned int element_count;
	unsigned int tex;
	const char *name;
	bool solid = false;
	bool invisible = false;
	bool dangerous = false;
	float radius = 1.0;
	vector<vec4> locations;

	GameObject(const char *n) : name(n) {}

	virtual void movement() {}

	bool init_from_object(const char *objfilename, vec3 (*input_transform)(float x, float y, float z) = 0) {
		tinyobj::attrib_t attrib;
		std::vector<tinyobj::shape_t> shapes;
		std::vector<tinyobj::material_t> materials;
		std::string warn, err;
		vector<Vertex> vertices;
		vector<uint32_t> indices;
		if(!tinyobj::LoadObj(&attrib, &shapes, &materials, &warn, &err, objfilename))
			throw std::runtime_error(err);

		unordered_map<Vertex, uint32_t> uniqueVertices = {};

		for(const auto& shape : shapes){
			for(const auto& index : shape.mesh.indices){
				Vertex vertex = {};
				if(input_transform){
					vertex.pos = input_transform(attrib.vertices[3 * index.vertex_index + 0],
						attrib.vertices[3 * index.vertex_index + 1],
						attrib.vertices[3 * index.vertex_index + 2]);
				} else {
					vertex.pos = {
						attrib.vertices[3 * index.vertex_index + 0],
						attrib.vertices[3 * index.vertex_index + 1],
						attrib.vertices[3 * index.vertex_index + 2]
					};	
				}	
				vertex.texCoord = {
					attrib.texcoords[2 * index.texcoord_index + 0],
					1.0 - attrib.texcoords[2 * index.texcoord_index + 1]
				};
				if(uniqueVertices.count(vertex) == 0){
					uniqueVertices[vertex] = (uint32_t)vertices.size();
					vertices.push_back(vertex);
				}
				indices.push_back(uniqueVertices[vertex]);
			}
		}

		float *non_padded_vertex = (float*)malloc(3 * sizeof(float) * vertices.size());
		float *non_padded_texture = (float*)malloc(2 * sizeof(float) * vertices.size());
		for(size_t i = 0; i < vertices.size(); i++){
			non_padded_vertex[i*3 + 0] = vertices[i].pos.x;
			non_padded_vertex[i*3 + 1] = vertices[i].pos.y;
			non_padded_vertex[i*3 + 2] = vertices[i].pos.z;
			non_padded_texture[i*2 + 0] = vertices[i].texCoord.x;
			non_padded_texture[i*2 + 1] = vertices[i].texCoord.y;
		}

		glGenBuffers(1, &vbuf);
		glBindBuffer(GL_SHADER_STORAGE_BUFFER, vbuf);
		glBufferData(GL_SHADER_STORAGE_BUFFER, 3 * sizeof(float) * vertices.size(), non_padded_vertex, GL_STATIC_DRAW); 
		free(non_padded_vertex);
		std::cout << "Loaded " << vertices.size() << " Vertices\n";

		glGenBuffers(1, &tbuf);
		glBindBuffer(GL_SHADER_STORAGE_BUFFER, tbuf);
		glBufferData(GL_SHADER_STORAGE_BUFFER, 2 * sizeof(float) * vertices.size(), non_padded_texture, GL_STATIC_DRAW); 
		free(non_padded_texture);

		glGenBuffers(1, &ebuf);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebuf);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned int) * indices.size(), indices.data(), GL_STATIC_DRAW);
		element_count = indices.size();
		std::cout << "Loaded " << indices.size() << " Indexes\n";

		return true;
	}

	bool in_contact_radial_location(float x, float y, float z, int idx){
		float dx = locations[idx].x - x;
		float dy = locations[idx].y - y;
		float dz = locations[idx].z - z;
		return (dx * dx + dy * dy + dz * dz) < (radius * radius);
	}

	virtual bool in_contact_location(float x, float y, float z, int i) {
		return in_contact_radial_location(x, y, z, i);
	}
	
	int in_contact(float x, float y, float z){
		for(int i = 0; i < locations.size(); i++){
			if(in_contact_location(x, y, z, i))
				return i;
		}
		return -1;
	}

	virtual bool init() = 0;

	// drawing:  set up and call glDrawElementsInstanced
	void draw_object(const mat4 &vp)  {
		glUseProgram(program);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, tex);
		glUniform1i(tex_uniform, 0);

		glUniformMatrix4fv(mvp_uniform, 1, 0, glm::value_ptr(vp));

		glEnableVertexAttribArray(vertex_attrib);
		glBindBuffer(GL_ARRAY_BUFFER, vbuf);
		glVertexAttribPointer(vertex_attrib, 3, GL_FLOAT, GL_FALSE, 0, 0);

		glEnableVertexAttribArray(tex_attrib);
		glBindBuffer(GL_ARRAY_BUFFER, tbuf);
		glVertexAttribPointer(tex_attrib, 2, GL_FLOAT, GL_FALSE, 0, 0);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebuf);
		glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, lbuf);
		glDrawElementsInstanced(GL_TRIANGLES, element_count, GL_UNSIGNED_INT, 0, locations.size());
	}

	bool setup_shaders(const char *vertex, const char *fragment){
		vertex_shader = make_shader(vertex, GL_VERTEX_SHADER);
		fragment_shader = make_shader(fragment, GL_FRAGMENT_SHADER);
		if(!(vertex_shader && fragment_shader))
			return 0;

		program = glCreateProgram();
		glAttachShader(program, vertex_shader);
		glAttachShader(program, fragment_shader);
		glLinkProgram(program);
		GLint link_ok;
		glGetProgramiv(program, GL_LINK_STATUS, &link_ok);
		if(!link_ok){
			glGetProgramInfoLog(program, GBLEN, NULL, general_buffer);
			puts(general_buffer);
			puts(RED("Link Failed").c_str());
			return false;
		}
	
	}

	bool setup_standard_shaders(){
		if(generic_program == -1){
			setup_shaders("cube_vertex.glsl", "cube_fragment.glsl");
			generic_program = program;
		} else 
			program = generic_program;
		vertex_attrib = glGetAttribLocation(program, "in_vertex");
		tex_attrib = glGetAttribLocation(program, "in_texcoord");
		mvp_uniform = glGetUniformLocation(program, "mvp");
		tex_uniform = glGetUniformLocation(program, "tex");
		return true;
	}

	bool setup_texture(const char *filename){
		int texwidth, texheight;
		unsigned char *image_data = SOIL_load_image(filename, &texwidth, &texheight, 0, SOIL_LOAD_RGB);
		printf("Loaded texture %d by %d\n", texwidth, texheight);
		if(!image_data)
			return false;
		glGenTextures(1, &tex);
		glBindTexture(GL_TEXTURE_2D, tex);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, texwidth, texheight, 0, GL_RGB, GL_UNSIGNED_BYTE, image_data);
		free(image_data);
		return true;
	}

	bool setup_locations(){
		glGenBuffers(1, &lbuf);
		update_locations();
		return true;
	}

	bool update_locations(){
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, lbuf);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(vec4) * locations.size(), locations.data(), GL_STATIC_DRAW);
		return true;
	}
	virtual void draw(mat4 vp) = 0;
};

struct Chalet : public GameObject {
	Chalet() : GameObject("Chalet") {}

	static vec3 input_transform(float x, float y, float z){
		return vec3(x*2, z*2, y*2);
	}

	bool init() override {
		init_from_object("chalet.obj", input_transform);
		setup_standard_shaders();
		setup_texture("chalet.jpg");
		setup_locations();
		return true;
	}
	void draw(mat4 vp){
		draw_object(vp);
	}
};
/*
struct ObjectFromFile : public GameObject {
	MonsterBox(char *n, char *objfile, char *texfile) : GameObject("MonsterBox") {}
	
}*/

struct GenericObject : public GameObject {
	const char *objfile;
	const char *texfile;
	GenericObject(const char *n, const char *of, const char *tf) : GameObject(n), objfile(of), texfile(tf) {}
	bool init() override {
		init_from_object(objfile);
		setup_standard_shaders();
		setup_texture(texfile);
		setup_locations();
		return true;
	}
	void draw(mat4 vp){
		draw_object(vp);
	}	
};

struct MovingGameObject : public GameObject {
	vector<vec4> velocity;
	MovingGameObject(const char *name) : GameObject(name) {}
	void addnew_radial(float x, float y, float z, float speed, float heading, float elevation){
		float vx = speed * cos(elevation) * cos(heading);
		float vy = speed * sin(elevation);
		float vz = speed * cos(elevation) * sin(heading);
		addnew_direct(x, y, z, vx, vy, vz);
	}
	void addnew_direct(float x, float y, float z, float vx, float vy, float vz){
		locations.push_back(vec4(x, y, z, 0));
		velocity.push_back(vec4(vx, vy, vz, 100));
	}
};

struct Projectile : public MovingGameObject {
	Projectile() : MovingGameObject("Projectile") {
		dangerous = true;
	}
	bool init() override {
		init_from_object("ball.obj");
		setup_standard_shaders();
		setup_texture("brick.jpg");
		setup_locations();
		printf("vertices.size() %d\n", element_count);
		return true;
	}
	void movement() override {
		// Update positions!
		for(int i = 0; i < locations.size(); i++){
			locations[i].x += velocity[i].x;
			locations[i].y += velocity[i].y;
			if(locations[i].y < -1)
				velocity[i].y *= -1;
			locations[i].z += velocity[i].z;
			velocity[i].w -= 1.0;
			if(velocity[i].w < 0.0){
				velocity[i] = velocity[velocity.size() - 1];
				velocity.pop_back();
				locations[i] = locations[locations.size() - 1];
				locations.pop_back();
			}
		}
	}
	void draw(mat4 vp){
		update_locations(); // copy new locations to video memory
		draw_object(vp);
	}
};

struct Fragment : public MovingGameObject {
	Fragment() : MovingGameObject("Fragment") {}
	bool init() override {
		init_from_object("ball.obj");
		setup_shaders("frag_vertex.glsl", "frag_fragment.glsl");
		setup_locations();
		return true;
	}
	void movement() override {
		// Update positions!
		for(int i = 0; i < locations.size(); i++){
			locations[i].x += velocity[i].x;
			locations[i].y += velocity[i].y;
			if(locations[i].y < -1)
				velocity[i].y *= -1;
			locations[i].z += velocity[i].z;
//			velocity[i].w -= 1.0;
			if(velocity[i].w < 0.0){
				velocity[i] = velocity[velocity.size() - 1];
				velocity.pop_back();
				locations[i] = locations[locations.size() - 1];
				locations.pop_back();
			}
			velocity[i].y -= 0.001;
		}
	}
	void draw(mat4 vp) {
		update_locations(); // copy new locations to video memory
		glUseProgram(program);
		glUniformMatrix4fv(mvp_uniform, 1, 0, glm::value_ptr(vp));

		glEnableVertexAttribArray(vertex_attrib);
		glBindBuffer(GL_ARRAY_BUFFER, vbuf);
		glVertexAttribPointer(vertex_attrib, 3, GL_FLOAT, GL_FALSE, 0, 0);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebuf);
		glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, lbuf);
		glDrawElementsInstanced(GL_TRIANGLES, element_count, GL_UNSIGNED_INT, 0, locations.size());
	}	
};

Fragment *fragments;
Projectile *projectiles;

struct MonsterBox : public GameObject {
	MonsterBox() : GameObject("MonsterBox"){
		dangerous = true;
	}
	bool location_refresh;
	bool init() override {
		init_from_object("cube.obj");
		setup_standard_shaders();
		setup_texture("monster1.png");
		setup_locations();
		location_refresh = false;
		return true;
	}
	void movement() override {
		for(auto go : objects)
		for(auto pl : projectiles->locations)
			for(int i = 0; i < locations.size(); i++) {
				vec4 mbl = locations[i];
				if(fabs(pl.x - mbl.x) < 0.5 && fabs(pl.y - mbl.y) < 0.5 && fabs(pl.z - mbl.z) < 0.5){
					destroy(i, mbl);
				}

			}
	}
	float random_velocity(){
		return (random() % 100 / 100.0f - 0.5f) * .2;
	}
	void destroy(int idx, vec4 mbl){
		for(int i = 0; i < 20; i++){

			fragments->addnew_direct(mbl.x, mbl.y, mbl.z, random_velocity(), random_velocity(), random_velocity());
		}
		locations[idx] = locations[locations.size() - 1];
		locations.pop_back();
		location_refresh = true;
	}
	void draw(mat4 vp){
		if(location_refresh){
			update_locations();
			location_refresh = false;
		}
		draw_object(vp);
	}
};

struct StoneBlock : public GameObject {
	StoneBlock() : GameObject("StoneBlock") {}
	bool location_refresh;
	// initialization:  Texture and vertices into memory
	bool init(){
		GLfloat cube_vertices[] = {
			// front
			-0.25, -0.25,  0.25,
			0.25, -0.25,  0.25,
			0.25,  0.25,  0.25,
			-0.25,  0.25,  0.25,
			// top
			-0.25,  0.25,  0.25,
			0.25,  0.25,  0.25,
			0.25,  0.25, -0.25,
			-0.25,  0.25, -0.25,
			// back
			0.25, -0.25, -0.25,
			-0.25, -0.25, -0.25,
			-0.25,  0.25, -0.25,
			0.25,  0.25, -0.25,
			// bottom
			-0.25, -0.25, -0.25,
			0.25, -0.25, -0.25,
			0.25, -0.25,  0.25,
			-0.25, -0.25,  0.25,
			// left
			-0.25, -0.25, -0.25,
			-0.25, -0.25,  0.25,
			-0.25,  0.25,  0.25,
			-0.25,  0.25, -0.25,
			// right
			0.25, -0.25,  0.25,
			0.25, -0.25, -0.25,
			0.25,  0.25, -0.25,
			0.25,  0.25,  0.25,
		};

		glGenBuffers(1, &vbuf);
		glBindBuffer(GL_SHADER_STORAGE_BUFFER, vbuf);
		glBufferData(GL_SHADER_STORAGE_BUFFER, sizeof(cube_vertices), cube_vertices, GL_STATIC_DRAW);

		GLushort cube_elements[] = {
			// front
			0,  1,  2,
			2,  3,  0,
			// top
			4,  5,  6,
			6,  7,  4,
			// back
			8,  9, 10,
			10, 11,  8,
			// bottom
			12, 13, 14,
			14, 15, 12,
			// left
			16, 17, 18,
			18, 19, 16,
			// right
			20, 21, 22,
			22, 23, 20,
		};
		element_count = sizeof(cube_elements) / sizeof(GLushort);
		glGenBuffers(1, &ebuf);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebuf);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(cube_elements), cube_elements, GL_STATIC_DRAW);

		GLfloat cube_texcoords[2*4*6] = {
			// front
			0.0, 0.0,
			1.0, 0.0,
			1.0, 1.0,
			0.0, 1.0,
		};
		for (int i = 1; i < 6; i++)
			memcpy(&cube_texcoords[i*4*2], &cube_texcoords[0], 2*4*sizeof(GLfloat));
		glGenBuffers(1, &tbuf);
		glBindBuffer(GL_ARRAY_BUFFER, tbuf);
		glBufferData(GL_ARRAY_BUFFER, sizeof(cube_texcoords), cube_texcoords, GL_STATIC_DRAW);

		setup_locations();
		setup_texture("brick.jpg");
		setup_standard_shaders();

		return true;
	}


	bool in_contact_location(float x, float y, float z, int i){
		const float margin = 0.1;
		if(abs(x - locations[i].x) < 0.25 + margin && abs(y - locations[i].y) < 0.25 + margin && abs(z - locations[i].z) < 0.25 + margin){
			return i;
		}
	}

	void draw(mat4 vp)  {
		if(location_refresh){
			update_locations();
			location_refresh = false;
		}
		glUseProgram(program);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, tex);
		glUniform1i(tex_uniform, 0);
		glUniformMatrix4fv(mvp_uniform, 1, 0, glm::value_ptr(vp));

		glEnableVertexAttribArray(vertex_attrib);
		glBindBuffer(GL_ARRAY_BUFFER, vbuf);
		glVertexAttribPointer(vertex_attrib, 3, GL_FLOAT, GL_FALSE, 0, 0);

		glEnableVertexAttribArray(tex_attrib);
		glBindBuffer(GL_ARRAY_BUFFER, tbuf);
		glVertexAttribPointer(tex_attrib, 2, GL_FLOAT, GL_FALSE, 0, 0);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebuf);
		glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, lbuf);
		glDrawElementsInstanced(GL_TRIANGLES, element_count, GL_UNSIGNED_SHORT, 0, locations.size());
	}

	// movement:  for falling calculations
	// Returns the effective floor level
	float find_floor(){
		float efloor = -1000;
		for(vec3 l : locations){
			if(abs(player_x - l.x) < 0.25 && abs(player_z - l.z) < 0.25){
				if(l.y < player_y + 2.0 && l.y > efloor)
					efloor = l.y + 1;
			}
		}
		return efloor;
	}

	void movement() override {
		for(auto pl : projectiles->locations)
			for(int i = 0; i < locations.size(); i++) {
				vec4 mbl = locations[i];
				if(fabs(pl.x - mbl.x) < 0.5 && fabs(pl.y - mbl.y) < 0.5 && fabs(pl.z - mbl.z) < 0.5){
					locations[i] = locations[locations.size() - 1];
					locations.pop_back();
					location_refresh = true;
				}

			}
	}
};
